/* Implementation File for the SeismicData Class */

#import "SeismicData.h"
#import <sys/ioctl.h>
#import <stdlib.h>
#import <string.h>
#import <math.h>
#import <string.h>


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


@implementation SeismicData


- init:sender
{
	[super init];
	[self nullTheHeader];
	data=NULL;
	return self;
}

- setFilePointer: (FILE *)ipt;
{
	pt=ipt;
	return self;
}

- (char *)getAhComment;
{
	return header.event.ecomment;
}

- getAhData:(char *)dt andHeader:(ahhed *)ahhead
{
	int num_bytes;
	
	num_bytes=sizeof(header);
	memcpy((char *)ahhead,(char *)&header,num_bytes);
	num_bytes=[self size]*header.record.ndata;
	memcpy(dt,(char *)data,num_bytes);
	return self;
}

- getLength:(int *)olength
{
	*olength=header.record.ndata;
	return self;
}

- getAhData:(char *)dt
{
	int num_bytes;
	num_bytes=[self size]*header.record.ndata;
	memcpy(dt,(char *)data,num_bytes);
	return self;
}	

- putAhData:(char *)dt andHeader:(ahhed *)ahhead
{
	int num_bytes;
	
	if(data != NULL)
	{
		free((void *)data);
		[self nullTheHeader];
	}
	[self makeDataSpace];
	num_bytes=sizeof(ahhed);
	memcpy((char *)&header,(char *)ahhead,num_bytes);
	num_bytes=size(ahhead)*ahhead->record.ndata;
	memcpy((char *)data,dt,num_bytes);
	return self;
}

- readAhRecord:(int)num
{
	id error;
	
	if(data != NULL)
	{
		free((void *)data);
		[self nullTheHeader];
	}
	if(num != 0)		/* if num=0 read file sequentially */
	{
		error=[self positionFile:num];
		if(error == nil) return nil;
	}
	error=[self readAhHeadFromFile];
	if(error == nil) return nil;
	[self makeDataSpace];
	error=[self readAhDataFromFile];
	if(error == nil)return nil;
	[self setMaxAmp];
	
	return self;
}

- writeAhRecord
{
	[self putAhHeadToFile];
	[self putAhDataToFile];
	
	return self;
}

- getDataType:(int *)type
{
	*type=header.record.type;
	return self;
}

- getDataSegment:(int)segment ofLength:(int *)length andSampleInterval:(float *)dt
	withMaxValue:(float *)maxvalue forPlot:(float **)plot_data;
{	
	int count;
	float *tfp;
	
	*maxvalue=0.0;
	*dt=header.record.delta;
	*length=header.record.ndata;
	
	if(*plot_data != NULL) free((void *)*plot_data);
	*plot_data=(float *)calloc((unsigned)(header.record.ndata),(unsigned)sizeof(float));
	tfp= *plot_data;
	
	if(header.record.type == FLOAT)
	{
		for(count=0;count<header.record.ndata;count++)
		{
			tfp[count]=data[count];
			if(fabs(tfp[count]) > *maxvalue) 
				*maxvalue=fabs(tfp[count]);
		}
		return self;
	}
	if(header.record.type == COMPLEX)
	{
		if(segment == REAL)
			for(count=0;count<header.record.ndata;count++)
			{
				tfp[count]=data[count*2];
				if(fabs(tfp[count]) > *maxvalue)
					 *maxvalue=fabs(tfp[count]);
			}
		else
			for(count=0;count<header.record.ndata;count++)
			{
				tfp[count]=data[count*2+1];
				if(fabs(tfp[count]) > *maxvalue)
					 *maxvalue=fabs(tfp[count]);
			}
		return self;
	}
	
	return nil;
}

- scaleData:(float)scale
{
	int i;
	
	[self setMaxAmp];
	for(i=0;i<header.record.ndata;i++)
		data[i]=(data[i]/header.record.maxamp)*scale;
		
	return self;
}

- fourier
{
	float *ftpf,amp,phase,nyquist,nf,freq_inc;
	int i,ndata;
	
		/* get nearest power of two to the listed data length */
	for(i=1;(int)pow((double)2,(double)i)<header.record.ndata;++i);
        ndata=(int)pow((double)2,(double)i);
	
		/* allocate space */
	ftpf=(float *)calloc((unsigned)(ndata+2),sizeof(float));
	
		/* initialize input array */
	for(i=0;i<ndata+2;i++)
		ftpf[i]=0.0;
	for(i=0;i<header.record.ndata;i++)
		ftpf[i]=data[i];
		
		/* transform array */
	cfftr(ftpf,ndata);
	
		/* convert output to amplitude and phase */
	for(i=0;i<=ndata;i+=2)
	{
		amp=sqrt((double)(ftpf[i]*ftpf[i]+ftpf[i+1]*ftpf[i+1]));
		phase=atan2(ftpf[i+1],ftpf[i]);
		ftpf[i]=amp*(2.0/ndata);
		ftpf[i+1]=phase;
	}
	
		/* free input data array */
	free((void *)data);
	
		/* assign data pointer to ftpf */
	data=ftpf;
	
		/* change other instance variables */
	header.record.type=COMPLEX;
	header.record.ndata=(int)(ndata/2.0+1.0);
	
        nyquist=1.0/(2.0*header.record.delta);
        nf=(int)(ndata/2.0+1.0);
        freq_inc=nyquist/(float)(nf-1);
        header.record.delta=freq_inc;
	
	
	return self;
}

		/* internal methods */
	
- setMaxAmp;
{
        float   *fpt;
        double *dpt, dmin, dmax;
        float   max,min;
        long    n_data_pts;

        switch(header.record.type)
        {
        	case FLOAT:
               	 	n_data_pts= header.record.ndata;
               	 	break;
        	case COMPLEX:
        	case VECTOR:
                	n_data_pts= 2 * header.record.ndata;
                	break;
        	case TENSOR:
                	n_data_pts= 3 * header.record.ndata;
                	break;
        	case 5:
                	n_data_pts= 4 * header.record.ndata;
                	break;
        	case DOUBLE:
                	n_data_pts= header.record.ndata;
                	break;
        	default:
                	return nil;
                break;
        }

        if (header.record.type == DOUBLE) {
                dpt= (double *)data;
                dmax= dmin= *dpt;
                while(n_data_pts--)
                {
                        dmax= MAX(dmax,*dpt);
                        dmin= MIN(dmin,*dpt);
                        ++dpt;
                }
                ((fabs(dmax) > fabs(dmin)) ? (header.record.maxamp= (float) dmax) : (header.record.maxamp= (float) -dmin));
        }
        else {
                fpt= (float *)data;
                max= min= *fpt;
                while(n_data_pts--)
                {
                        max= MAX(max,*fpt);
                        min= MIN(min,*fpt);
                        ++fpt;
                }
                ((fabs((double)max) > fabs((double)min)) ? (header.record.maxamp= max) : (header.record.maxamp= -min));
        }

        return self;
}


- readAhHeadFromFile
{
        int     ierr = 0;

        if((ierr = fread((char *)&header,sizeof(ahhed),1,pt)) == 1)
        {
                if((header.record.type < TYPEMIN) || (header.record.type > TYPEMAX))
                {
                        [self nullTheHeader];
                        return nil;
                }
        }
        else            /* not enough head */
        {
                [self nullTheHeader];
		return nil;
        }
        return self;
}

- putAhHeadToFile
{
	int ierr;
	
        if((ierr= fwrite((char *)&header,sizeof(ahhed),1,pt)) != 1)
        {
                return nil;
        }
        return self;
}

- (int)size;
{
        int     type_size = 0;

        switch(header.record.type)
        {
        case 1:         /* real time series */
                type_size= sizeof(float);
                break;
        case 2:         /* complex time series */
                type_size= sizeof(complex);
                break;
        case 3:         /* real x,y pairs */
                type_size= sizeof(vector);
                break;
        case 4:         /* x real, y complex, or real x,y,z */
                type_size= sizeof(tensor);
                break;
        case 5:         /* complex x,y pairs */
                type_size= 2*sizeof(complex);
                break;
        case 6:         /* double */
                type_size=sizeof(double);
                break;
        default:        /* unknown data type */
                type_size= -1;
		return 0;
        }
        return(type_size);
}

- positionFile:(int)n
{
        int     i;

        rewind(pt);
        for(i=1; i<n; ++i)
        {
                if([self readAhHeadFromFile] != nil)
                {
                        if(fseek(pt,(long)(header.record.ndata)*(size(&header)),1) == -1)
                        {
                                return nil;
                        }
                }
                else
                {
                        return nil;
                }
        }
	[self nullTheHeader];
        return self;      /* success */
}

- readAhDataFromFile
{
        int ierr = 0;

        if((ierr = fread(data,[self size],(int)header.record.ndata,pt)) != (int)header.record.ndata)
        {
                return nil;
        }

        return self;
}

- putAhDataToFile
{
        int     ierr = 0;

        if((ierr = fwrite(data,[self size],(int)header.record.ndata,pt)) != (int)header.record.ndata)
        {
                return nil;
        }
        return self;
}

- makeDataSpace
{
        
        data=(float *)calloc((unsigned)header.record.ndata,(unsigned)[self size]);
	return self;
}

-nullTheHeader
{
	int i;
	
        strcpy(header.station.code,"null");
        strcpy(header.station.chan,"null");
        strcpy(header.station.stype,"null");
        header.station.slat= 0.0;
        header.station.slon= 0.0;
        header.station.elev= 0.0;
        header.station.DS= 0.0;
        header.station.A0= 0.0;
        for(i=0; i< NOCALPTS; ++i)
        {
                header.station.cal[i].pole.r= 0.0;
                header.station.cal[i].pole.i= 0.0;
                header.station.cal[i].zero.r= 0.0;
                header.station.cal[i].zero.i= 0.0;
        }

        header.event.lat= 0.0;
        header.event.lon= 0.0;
        header.event.dep= 0.0;
        header.event.ot.yr= (short)0;
        header.event.ot.mo= (short)0;
        header.event.ot.day= (short)0;
        header.event.ot.hr= (short)0;
        header.event.ot.mn= (short)0;
        header.event.ot.sec= 0.0;
        strcpy(header.event.ecomment,"null");

        header.record.type= (short)0;
        header.record.ndata= 0L;
        header.record.delta= 0.0;
        header.record.maxamp= 0.0;
        header.record.abstime.yr= (short)0;
        header.record.abstime.mo= (short)0;
        header.record.abstime.day= (short)0;
        header.record.abstime.hr= (short)0;
        header.record.abstime.mn= (short)0;
        header.record.abstime.sec= 0.0;
        header.record.rmin= 0.0;
        strcpy(header.record.rcomment,"null");
        strcpy(header.record.log,"null");

        for(i=0; i< NEXTRAS; ++i)
                header.extra[i]= 0.0;

        return self;
}


@end


	/* ah ioroutines used by the methods implemented in this class */
	
/*      size
 *              returns the size (in bytes) of the data type given by
 *              head->record.type.
 *      returns:
 *                      size of data type       ->      no error
 *                      -1                      ->      unknown data type
 */
int     size(head)
        ahhed   *head;
{
        int     type_size = 0;

        switch(head->record.type)
        {
        case 1:         /* real time series */
                type_size= sizeof(float);
                break;
        case 2:         /* complex time series */
                type_size= sizeof(complex);
                break;
        case 3:         /* real x,y pairs */
                type_size= sizeof(vector);
                break;
        case 4:         /* x real, y complex, or real x,y,z */
                type_size= sizeof(tensor);
                break;
        case 5:         /* complex x,y pairs */
                type_size= 2*sizeof(complex);
                break;
        case 6:         /* double */
                type_size=sizeof(double);
                break;
        default:        /* unknown data type */
                type_size= -1;
                break;
        }
        return(type_size);
}


void cfftr(float *x,int n )
{
        int nn, is, nm, j, i;
        int k1j, k1i, k2j, k2i;
        float s, fn, ex, wr, wi, wwr, wrr, wwi, a1, a2, b1, b2;
        double sin(), cos();
        nn = n/2;
        is = 1;
        cfour( x, nn, is );
        nm = nn/2;
        s = x[0];
        x[0] += x[1];
        x[n] = s - x[1];
        x[1] = 0.0 ;
        x[n+1] = 0.0;
        x[nn+1] = (-x[nn+1]);
        fn = (float) n;
        ex = 6.2831853 / fn;
        j = nn;
        wr = 1.0;
        wi = 0.0;
        wwr = (float) cos( (double) ex );
        wwi = (float) (-sin( (double) ex ));
        for (i=2; i<=nm; i++) {
                wrr = wr*wwr-wi*wwi;
                wi = wr*wwi+wi*wwr;
                wr = wrr;
                k1j = 2*j-1;
                k1i = 2*i-1;
                k2j = 2*j;
                k2i = 2*i;
                a1 = 0.5*(x[k1i-1]+x[k1j-1]);
                a2 = 0.5*(x[k2i-1]-x[k2j-1]);
                b1 = 0.5*(-x[k1i-1]+x[k1j-1]);
                b2 = 0.5*(-x[k2i-1]-x[k2j-1]);
                s = b1;
                b1 = b1*wr+b2*wi;
                b2 = b2*wr-s*wi;
                x[k1i-1] = a1-b2;
                x[k2i-1] = (-a2-b1);
                x[k1j-1] = a1+b2;
                x[k2j-1] =  a2-b1;
                j -= 1;
        }
        return;
}

void cfour(float *data,int n,int isign )
{
        int ip0, ip1, ip2, ip3, i3rev;
        int i1, i2a, i2b, i3;
        float sinth, wstpr, wstpi, wr, wi, tempr, tempi, theta;
        double sin();
        ip0=2;
        ip3=ip0*n;
        i3rev=1;
        for( i3=1; i3<=ip3; i3+=ip0 ) {
                if( i3 < i3rev ) {
                        tempr = data[i3-1];
                        tempi = data[i3];
                        data[i3-1] = data[i3rev-1];
                        data[i3] = data[i3rev];
                        data[i3rev-1] = tempr;
                        data[i3rev] = tempi;
                        }
                ip1 = ip3 / 2;
                do {
                        if( i3rev <= ip1 )
                                break;
                        i3rev -= ip1;
                        ip1 /= 2;
                        }
                        while ( ip1 >= ip0 );
                i3rev += ip1;
                }
                ip1 = ip0;
                while ( ip1 < ip3 ) {
                        ip2 = ip1 * 2;
                        theta = 6.283185/( (float) (isign*ip2/ip0) );
                        sinth = (float) sin( (double) (theta/2.) );
                        wstpr = -2.*sinth*sinth;
                        wstpi = (float) sin( (double) theta );
                        wr = 1.;
                        wi = 0.;
                        for ( i1=1; i1<=ip1; i1+=ip0 ) {
                        for ( i3=i1; i3<ip3; i3+=ip2 ) {
                                i2a=i3;
                                i2b=i2a+ip1;
                                tempr = wr*data[i2b-1] - wi*data[i2b];
                                tempi = wr*data[i2b] + wi*data[i2b-1];
                                data[i2b-1] = data[i2a-1] - tempr;
                                data[i2b] = data[i2a] - tempi;
                                data[i2a-1] += tempr;
                                data[i2a] += tempi;
                                }
                                tempr = wr;
                                wr = wr*wstpr - wi*wstpi + wr;
                                wi = wi*wstpr + tempr*wstpi + wi;
                                }
                        ip1=ip2;
                        }
                return;
                }
