
/* Generated by Interface Builder */

#import "Controller.h"
#import "ModelControl.h"
#import "ParamControl.h"
#import "SpecControl.h"
#import "PsSpecControl.h"
#import "TraceView.h"

#import <appkit/Matrix.h>
#import <appkit/Button.h>
#import <appkit/ButtonCell.h>
#import <appkit/Cell.h>
#import <appkit/Application.h>


#import <stdlib.h>
#import <stdio.h>
#import <math.h>

#define PI	3.1415926
#define CONVRHO	1.0e+12	/* magic number to convert units on density */

typedef struct {
	float r;
	float i;
	} complex;

@implementation Controller


- appDidInit:sender
{
	id myButtonCell;
	
	[myModelControl prepareArea];
	[myInputControl prepareArea];
	[mySpecControl prepareArea];
	[myPsSpecControl prepareArea];

		/* initialize component matrix */
	[CompMatrix selectCellAt:trace :0];
	myButtonCell=[CompMatrix cellAt:2 :0];
	[myButtonCell setEnabled:NO];
	
		/* initialize type matrix */
	[TypeMatrix selectCellAt:trace :0];

	return self;
}
	/* action methods */
- setComponent:sender
{
	trace = [sender selectedRow];	
	[myTraceView display];
	[mySpecControl prepareArea];
	[myPsSpecControl prepareArea];

    	return self;
}

- showHelp:sender
{
    	if (!myHelpPanel) 
		[NXApp loadNibSection:"Help.nib" owner:self];
		
    	[myHelpPanel makeKeyAndOrderFront:NULL];
 
   	return self;
}

- setType:sender
{
	type = [sender selectedRow];
	[myCompButton performClick:self];

	return self;
}
- init
{
	vert=rad=tang=NULL;
	ndata=0;
	dt=0.0;
	
	vert=(float *)calloc(8194,sizeof(float));
	rad=(float *)calloc(8194,sizeof(float));
	tang=(float *)calloc(8194,sizeof(float));

	trace=VERT;
	type=DISP;
	
	return self;
}

	/* action methods */
- computeGroundMotion:sender
{
	float rayp,nyquist,nf,freq_inc,w,delta,scale,save,k;
	float *tarray;
	int i;
	
		/* function protypes */
	void PSVsurfacedisplace(float z[], float beta[], float alpha[], float rho[], float q[],
		int number_layers, float ray_p, complex urad[], complex uvert[], float nyquist,
		int data_length,char taper,char type);
	void SHsurfacedisplace(float z[], float beta[], float rho[], float q[],
		int number_layers, float ray_p, complex u[], float nyquist,
		int data_length,char taper);
	void t_fold(float *a,int ndata_a,float *b,int ndata_b,float *result);
	
	void cfftr(),cfftri();
	
		/* input parameters */
	float incAngle;
	int	itype;
	
		/* velocity model */
	float *t1,*t2,*t3,*t4,*t5;
	float *p,*s,*q,*rho,*z;
	int	numLayers;
	
		/* now query the model and parameter controllers for info */
	[myInputControl sendParameters:&dt:&incAngle:&ndata:&itype:&sTime:&lSTime];
	[myModelControl sendModel:&t1:&t2:&t3:&t4:&t5:&numLayers];
	p=t1-1;
	s=t2-1;
	q=t3-1;
	rho=t4-1;
	z=t5-1;
	
		/* setup tarray to do convolution of source time with output */
	tarray=(float *)calloc(ndata+lSTime,sizeof(float));
	
		/* now calculate the ray parameter and nyquist frequency */
	if(itype == P)
		rayp=sin(incAngle/57.29577951)/p[numLayers];
	else
		rayp=sin(incAngle/57.29577951)/s[numLayers];
	nyquist=1.0/(2.0*dt);
	
		/* compute ground motion */
	if(itype == P)	/* incident P-wave */
	{
		PSVsurfacedisplace(z,s,p,rho,q,numLayers,rayp,(complex *)rad,(complex *)vert,
			nyquist,ndata,'y','p');
	}
	if(itype == SV)	/* incident SV-wave */
	{
		PSVsurfacedisplace(z,s,p,rho,q,numLayers,rayp,(complex *)rad,(complex *)vert,
			nyquist,ndata,'y','s');	
	}
	if(itype == SH) /* incident SH-wave */
	{
		SHsurfacedisplace(z,s,rho,q,numLayers,rayp,(complex *)tang,
			nyquist,ndata,'y');
	}
	
			
	nyquist=1.0/(2.0*dt);
       	nf=(int)(ndata/2.0+1.0);
        freq_inc=nyquist/(float)(nf-1);

		/* convolve source-time function with output */
		/* and compute desired type of ground motion */
	t_fold(vert,ndata,sTime,lSTime,tarray);
	for(i=0;i<ndata;i++) 
	{
		vert[i]=tarray[i];
		tarray[i]=0.0;
	}
	
	/* if we want velocity or acceleration calculate it */
	/* spectra is multiplied by a tapered version of w */
	/* taper is as given in Seismic Migration by A. J. Berkhout */
	if(type == ACC)
	{
		cfftr(vert,ndata);
		delta=dt;
		for(i=0;i<=ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=-4.0*sin(k*delta/2.0)*sin(k*delta/2.0)/((1.0-sin(k*delta/2.0)*sin(k*delta/2.0)/3.0)*delta*delta);
			vert[i] *= scale;
			vert[i+1] *= scale;
		}

		cfftri(vert,ndata);
	}
	if(type == VEL)
	{
		cfftr(vert,ndata);
		delta=dt;
		for(i=0;i<ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=3.0*sin(k*delta)/(2.0*delta*(1.0+0.5*cos(k*delta)));
                        save=vert[i]*scale;
                        vert[i]= -1.0*vert[i+1]*scale;
                        vert[i+1]=save;

		}
		
		cfftri(vert,ndata);
	}

	
	t_fold(rad,ndata,sTime,lSTime,tarray);
	for(i=0;i<ndata;i++) 
	{
		rad[i]=tarray[i];
		tarray[i]=0.0;
	}
	
	/* if we want velocity or acceleration calculate it */
	/* spectra is multiplied by a tapered version of w */
	/* taper is as given in Seismic Migration by A. J. Berkhout */
	if(type == ACC)
	{
		cfftr(rad,ndata);
		delta=dt;
		for(i=0;i<=ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=-4.0*sin(k*delta/2.0)*sin(k*delta/2.0)/((1.0-sin(k*delta/2.0)*sin(k*delta/2.0)/3.0)*delta*delta);
			rad[i] *= scale;
			rad[i+1] *= scale;
		}
		cfftri(rad,ndata);
	}
	if(type == VEL)
	{
		cfftr(rad,ndata);
		delta=dt;
		for(i=0;i<ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=3.0*sin(k*delta)/(2.0*delta*(1.0+0.5*cos(k*delta)));
                        save=rad[i]*scale;
                        rad[i]= -1.0*rad[i+1]*scale;
                        rad[i+1]=save;

		}
		
		cfftri(rad,ndata);
	}

	t_fold(tang,ndata,sTime,lSTime,tarray);
	for(i=0;i<ndata;i++) 
	{
		tang[i]=tarray[i];
		tarray[i]=0.0;
	}
	
	/* if we want velocity or acceleration calculate it */
	/* spectra is multiplied by a tapered version of w */
	/* taper is as given in Seismic Migration by A. J. Berkhout */
	if(type == ACC)
	{
		cfftr(tang,ndata);
		delta=dt;
		for(i=0;i<=ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=-4.0*sin(k*delta/2.0)*sin(k*delta/2.0)/((1.0-sin(k*delta/2.0)*sin(k*delta/2.0)/3.0)*delta*delta);
			tang[i] *= scale;
			tang[i+1] *= scale;
		}
		cfftri(tang,ndata);
	}
	if(type == VEL)
	{
		cfftr(tang,ndata);
		delta=dt;
		for(i=0;i<ndata;i+=2)
		{
			k=(float)i/2.0*freq_inc;
			scale=3.0*sin(k*delta)/(2.0*delta*(1.0+0.5*cos(k*delta)));
                        save=tang[i]*scale;
                        tang[i]= -1.0*tang[i+1]*scale;
                        tang[i+1]=save;

		}
		
		cfftri(tang,ndata);
	}

		/* display waveform */
	[myTraceView setValues];
	[myTraceView display];
	
		/* display spectra */
	[mySpecControl prepareArea];
	
		/* display Building response */
	[myPsSpecControl prepareArea];

	free((void *)tarray);			
	return self;

}

	/* methods delegated from trace view */
- sendDataToPlot:sender:(float **)y:(int *)length:(float *)sample
{
	*length=ndata;
	*sample=dt;
	
	if(trace == VERT)
	{
		if(vert[0] == 0.0)
			*y=NULL;
		else
			*y=vert;
	}
	if(trace == RAD)
	{
		if(rad[0] == 0.0)
			*y=NULL;
		else
			*y=rad;
	}
	if(trace == TANG)
	{
		if(tang[0] == 0.0)
			*y=NULL;
		else
			*y=tang;
	}
		
	return self;
}

	/* methods delegated from speccontroller */
- sendDataToSpec:sender:(float **)y:(int *)length:(float *)sample
{
	*length=ndata;
	*sample=dt;
	if(trace == VERT)
	{
		if(vert[0] == 0.0)
			*y=NULL;
		else
			*y=vert;
	}
	if(trace == RAD)
	{
		if(rad[0] == 0.0)
			*y=NULL;
		else
			*y=rad;
	}
	if(trace == TANG)
	{
		if(tang[0] == 0.0)
			*y=NULL;
		else
			*y=tang;
	}
		
	return self;
}

	/* methods delegated from paramcontroller */
- disableComp:sender:(int)component
{
	id myButtonCell;
	
		/* disable selected component */
	myButtonCell=[CompMatrix cellAt:component :0];
	[myButtonCell setEnabled:NO];
	
	return self;
}

- enableComp:sender:(int)component
{
	id myButtonCell;
	
		/* enable selected component */
	myButtonCell=[CompMatrix cellAt:component :0];
	[myButtonCell setEnabled:YES];
	
		/* select newly enable component */
	[CompMatrix selectCellAt:component :0];
	trace=component;

	return self;
}

@end


	/* c routines to do the calculations */
	
/*
	Routines to do complex arithmetic
	From Numerical Recipes
*/

complex cadd(complex a, complex b)	/* Returns complex sum of a and b */
{
	complex c;
	
	c.r=a.r+b.r;
	c.i=a.i+b.i;
	
	return c;
}

complex csub(complex a, complex b)	/* Returns complex difference of a and b */
{
	complex c;
	
	c.r=a.r-b.r;
	c.i=a.i-b.i;
	
	return c;
}

complex cmul(complex a, complex b)	/* Returns complex product of a and b */
{
	complex c;
	
	c.r=a.r*b.r-a.i*b.i;
	c.i=a.i*b.r+a.r*b.i;
	
	return c;
}

complex cdiv(complex a, complex b)	/* Returns complex quotient of a and b */
{
	complex c;
	float r,den;
	
	if(fabs(b.r) >= fabs(b.i))
	{
		r=b.i/b.r;
		den=b.r+r*b.i;
		c.r=(a.r+r*a.i)/den;
		c.i=(a.i-r*a.r)/den;
	}
	else
	{
		r=b.r/b.i;
		den=b.i+r*b.r;
		c.r=(a.r*r+a.i)/den;
		c.i=(a.i*r-a.r)/den;
	}
	
	return c;
}

complex comp(float re, float im)	/* returns a complex number with the specified parts */
{
	complex c;
	
	c.r=re;
	c.i=im;
	
	return c;
}

float c_abs(complex z)	/* Returns the absolute value of z */
{
	float x,y,ans,temp;
	
	x=fabs(z.r);
	y=fabs(z.i);
	if(x== 0.0)
	{
		ans=y;
	}
	else
	{
		if(y == 0.0)
		{
			ans=x;
		}
		else
		{
			if(x > y)
			{
				temp=y/x;
				ans=x*sqrt(1.0+temp*temp);
			}
			else
			{
				temp=x/y;
				ans=y*sqrt(1.0+temp*temp);
			}
		}
	}
	
	return ans;
}

complex conjg(complex z)	/* Returns the complex conjugate of z */
{
	complex c;
	
	c.r=z.r;
	c.i= -1.0*z.i;
	
	return c;
}

complex csqrt(complex z)	/* Returns the complex square root of z */
{
	complex c;
	float x,y,w,r;
	
	if(z.r == 0.0 && z.i == 0.0)
	{
		c.r=c.i=0.0;
		return c;
	}
	else
	{
		x=fabs(z.r);
		y=fabs(z.i);
		if( x >= y)
		{
			r=y/x;
			w=sqrt(x)*sqrt(0.5*(1.0+sqrt(1.0+r*r)));
		}
		else
		{
			r=x/y;
			w=sqrt(y)*sqrt(0.5*(r+sqrt(1.0+r*r)));
		}
		if(z.r >= 0.0)
		{
			c.r=w;
			c.i=z.i/(2.0*w);
		}
		else
		{
			c.i=(z.i >=0) ? w : -w;
			c.r=z.i/(2.0*c.i);
		}
		return c;
	}
}

complex rcmul(float x, complex a)	/* Returns the product of a float and a complex */
{
	complex c;
	
	c.r=x*a.r;
	c.i=x*a.i;
	
	return c;
}

float camp(complex a)	/* compute the amplitude of a */
{
	float amplitude;
	
	amplitude=sqrt((a.r*a.r+a.i*a.i));
	return(amplitude);
}

complex cexp(complex a)	/* compute e ^ a */
{
	complex c,e;
	float d;
	
	/* assume a is complex -> a=b+ic e^a=e^be^ic */
	
	d=exp(a.r);
			/* check for overflow */
	if(d == HUGE_VAL)
	{
		fprintf(stderr,"Error in cexp -> overflow\n");
		exit(-1);
	}
	e=comp( cos(a.i), sin(a.i) );
	c=rcmul(d,e);
	
	return c;
}

/* 
	Routines to do complex matrix manipulation
		You need to also include the complex arithmatic routines
*/

float mscale(complex a[],int m,int n)
/*
	This routine calculates a scale factor for the matrix a[] and
	then scales the matrix by that factor.
*/
{
	float sfact,test;
	int i,j;
	
	sfact=0.0;
	
	for(i=0;i<m;i++)
	{
		for(j=0;j<n;j++)
		{
			test=camp(a[i*n+j]);
			if(test > sfact)sfact=test;
		}
	}
	sfact=1.0/sfact;
	
	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			a[i*n+j]=rcmul(sfact,a[i*n+j]);
			
	return(sfact);
}

void cmat_mult(complex a[],int m,int n,complex b[],int p,complex c[])
/*
	This routine multiplies an m x n matrix a by an n x p matrix b
	and places the result in c
*/
{
	int i,j,k;
	complex temp;
	complex cmul(),cadd();
	
	for(i=0;i<m;i++)
	{
		for(j=0;j<p;j++)
		{
			c[i*p+j].r=c[i*p+j].i=0.0;
			for(k=0;k<n;k++)
			{
				temp=cmul(a[i*n+k],b[k*p+j]);
				c[i*p+j]=cadd(temp,c[i*p+j]);
			}
		}
	}
	return;
}

void cmat_trans(complex a[],int m,int n,complex b[])
/*
	This routine takes an m x n matrix a and constructs its transpose
	in b
*/
{
	int i,j;

	for(i=0;i<m;i++)
		for(j=0;j<n;j++)
			b[j*m+i]=a[i*n+j];
	return;
}

/*
	This routine calculates the ground displacement observed from an
	P-SV wave impinging from below a stack of layers

	Propagator matrices are used, Notation is as in Aki and Richards,
	Chapter 5.
	
	Need Subroutines;
	
		makeF_PSV -> Calculates the Layer Matrix F for each layer at
					one frequency
		makeFinv_PSV	-> Calculates the inverse of the Layer Matrix F for
						each layer at one frequency
		cmat_mult -> Multiplies two complex matrices (this calls routines
					to perform complex arithmetic
*/

void PSVsurfacedisplace(float z[], float beta[], float alpha[], float rho[], float q[],
	int number_layers, float ray_p, complex urad[], complex uvert[], float nyquist,
	int data_length,char taper,char type)
{
	complex eta[MAXLAYERS+1];
	complex xi[MAXLAYERS+1];
	complex pvel[MAXLAYERS+1],svel[MAXLAYERS+1];
	
				/* alpha, beta, rho, q, eta, xi, and z define the properties of each layer */
				/* alpha -> p wave velocity */
				/* beta -> shear wave velocity */
				/* q -> Q in each layer */
				/*	rho -> density */
				/* z -> the depth to the bottom of each layer */
				/*		the depth to the last layer is not used */
				/*		last layer is assumed to be a half space */
				/* eta -> cosj/beta in each layer */
				/* xi -> cosi/alpha in each layer */
				/* pvel, svel -> complex velocities in each layer */
				/* NOTE all of these values start from the shallowest */
				/* layer to the deepest.  First layer is stored at */
				/* an index of 1.  The index zero is the free surface */
				
				/* urad -> time domain radial displacment at the suface */
				/* uvert -> time domain vertical displacement at the surface */
				/* ray_p -> ray parameter of incoming SH wave */
				/* nyquist -> nyquist frequency to be used */
				/* data_length -> length of the output time series */
				/*	must be a factor of 2 */
				/* taper -> if 'y' spectra is tapered in the frequency */
				/*	domain with a cosine taper */
				/* type -> 'p' or 's' describes incident wave type */

				
	complex F[4][4],Finv[4][4];	/* layer matrices See Aki and Richards */
	complex S_down,P_down;		/* P and S waves coming off the bottom of the stack */
	complex temp[4][4],temp1[4][4],temp2[4][4],n2; /* work space */
	complex denominator,numerator;
	int i,j,count,nf,layer;
	float w,dw,freq_inc,zref,scale,sfact;

				/* Prototype all called functions */
				
	void makeF_PSV(float rho,float p,complex alpha,complex beta,complex xi,complex eta,float zref,float z,float w, 
		complex F[4][4]);
	void makeFinv_PSV(float rho,float p,complex alpha,complex beta,complex xi,complex eta,float zref,float z,float w, 
		complex F[4][4]);
	void cmat_mult(complex *F,int m, int n, complex *Finv,int p, complex *temp1);
	void cfftri(float *a,int dlength);
	complex comp(float r,float i);	
	complex cdiv(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cmul(complex a,complex b);
	complex cadd(complex a,complex b);
	complex csqrt(complex a);
	complex conjg(complex a);
	float dscale(complex *a,int m,int n);	
	
		
		/* set reference depth to depth of bottom of the stack */
	
	zref=z[number_layers-1];
	z[0]=0.0;

		/* calculate frequency increment and the number of frequencies */
		
	nf=(int)(data_length/2.0+1.0);
	freq_inc=nyquist/(float)(nf-1.0);
	dw=freq_inc*2.0*PI;
	
			/* loop over each frequency and calculate ground response */
	count=0;
	for(w=0.0;w < nyquist*2.0*PI+dw;w+=dw)
	{
				/* reset at temporary storage array */
				/* to the identity matrix */
				
		temp[0][0]=temp[1][1]=temp[2][2]=temp[3][3]=comp(1.0,0.0);
		temp[0][1]=temp[0][2]=temp[0][3]=comp(0.0,0.0);
		temp[1][0]=temp[1][2]=temp[1][3]=comp(0.0,0.0);
		temp[2][0]=temp[2][1]=temp[2][3]=comp(0.0,0.0);
		temp[3][0]=temp[3][1]=temp[3][2]=comp(0.0,0.0);
		
		if(w == 0.0)w+=(dw/100.0);	/* set w to something small */
		sfact=1.0;
		
			/* calculate eta, xi, p, and s velocities in each layer */
		
		for(i=number_layers;i>0;i--)
		{
			svel[i].r=beta[i]*(1.0+log(w/6.28318531)*1.0/(3.1415926*q[i]));
			svel[i].i= -1.0*beta[i]/(2.0*q[i]);
			pvel[i].r=alpha[i]*(1.0+log(w/6.28318531)*1.0/(3.1415926*q[i]));
			pvel[i].i= -1.0*alpha[i]/(2.0*q[i]);
			
			n2=cmul(svel[i],svel[i]);
			n2=cdiv(comp(1.0,0.0),n2);
			n2=cadd(comp((-1.0*ray_p*ray_p),0.0),n2);
			eta[i]=csqrt(n2);
			
			n2=cmul(pvel[i],pvel[i]);
			n2=cdiv(comp(1.0,0.0),n2);
			n2=cadd(comp((-1.0*ray_p*ray_p),0.0),n2);
			xi[i]=csqrt(n2);
		}
			
			/* construct propagator matrices for each layer */
				
		for(layer=0;layer < number_layers; layer++)
		{
				/* for every layer but the half space get F*Finv */
					
			if(layer != number_layers-1)
			{
				makeF_PSV(rho[layer+1]*CONVRHO,ray_p,pvel[layer+1],svel[layer+1],xi[layer+1],
					eta[layer+1],zref,z[layer],w,F);
				makeFinv_PSV(rho[layer+1]*CONVRHO,ray_p,pvel[layer+1],svel[layer+1],xi[layer+1],
					eta[layer+1],zref,z[layer+1],w,Finv);
				cmat_mult((complex *)F,4,4,(complex *)Finv,4,(complex *)temp1);
				/*sfact *= mscale((complex *)temp1,4,4);*/
					
						/* accumulate these over all layers */
							
				cmat_mult((complex *)temp,4,4,(complex *)temp1,4,(complex *)temp2);
				for(i=0;i<4;i++)
					for(j=0;j<4;j++)
						temp[i][j]=temp2[i][j];
			}
			else	/* in half space only calculate F */
			{
				makeF_PSV(rho[layer+1]*CONVRHO,ray_p,pvel[layer+1],svel[layer+1],xi[layer+1],
					eta[layer+1],zref,z[layer],w,F);
				cmat_mult((complex *)temp,4,4,(complex *)F,4,(complex *)temp1);
				for(i=0;i<4;i++)
					for(j=0;j<4;j++)
						temp[i][j]=temp1[i][j];
			}
		}
				/* temp now contains the contribution of each layer */
				/* at frequency w */
				/* now solve the matrix equation */
				/* in the follow t00 = temp[0][0] */
				/*
					Ux		t00 t01 t02 t03		P-down
					Uz		t10 t11 t12 t13		S-down
						=
					0		t20 t21 t22 t23		P-up
					0		t30 t31 t32 t33		S-up
				*/
				/*	where Ux is the observed radial ground displacement */
				/*	Uz is the observed vertial ground displacement */
				/* 	S-down is the reflected S wave coming */
				/*	out of the bottom of the stack */
				/*	P-down is the reflected P wave coming */
				/*	out of the bottom of the stack */
				/*	P-up and S-up are the incident P and S waves onto the */
				/*	bottom of the stack - one of these is 1 the other is 0 */
				/*	depending on whether we want and incident P wave or S wave */
			
		if(type == 'p')	/* incident wave is a P-wave, therefore S_up=0 */
		{
			denominator=cdiv(cmul(temp[2][0],temp[3][1]),temp[3][0]);
			denominator=rcmul(-1.0,denominator);
			denominator=cadd(temp[2][1],denominator);
			numerator=cdiv(cmul(temp[2][0],temp[3][2]),temp[3][0]);
			numerator=cadd(numerator,rcmul(-1.0,temp[2][2]));
			S_down=cdiv(numerator,denominator);
			numerator=cmul(S_down,temp[3][1]);
			numerator=rcmul(-1.0,cadd(temp[3][2],numerator));
			P_down=cdiv(numerator,temp[3][0]);
			
			urad[count]=cadd(cmul(P_down,temp[0][0]),cmul(S_down,temp[0][1]));
			urad[count]=cadd(temp[0][2],urad[count]);
			urad[count]=rcmul((1.0/sfact),urad[count]);
			uvert[count]=cadd(cmul(P_down,temp[1][0]),cmul(S_down,temp[1][1]));
			uvert[count]=cadd(temp[1][2],uvert[count]);
			uvert[count]=rcmul((1.0/sfact),uvert[count]);
		}	
		else	/* incident wave is an S-wave, therefore P_up=0 */
		{
			denominator=cdiv(cmul(temp[2][0],temp[3][1]),temp[3][0]);
			denominator=rcmul(-1.0,denominator);
			denominator=cadd(temp[2][1],denominator);
			numerator=cdiv(cmul(temp[2][0],temp[3][3]),temp[3][0]);
			numerator=cadd(numerator,rcmul(-1.0,temp[2][3]));
			S_down=cdiv(numerator,denominator);
			numerator=cmul(S_down,temp[3][1]);
			numerator=rcmul(-1.0,cadd(temp[3][3],numerator));
			P_down=cdiv(numerator,temp[3][0]);
	
			urad[count]=cadd(cmul(P_down,temp[0][0]),cmul(S_down,temp[0][1]));
			urad[count]=cadd(temp[0][3],urad[count]);
			urad[count]=rcmul((1.0/sfact),urad[count]);
			uvert[count]=cadd(cmul(P_down,temp[1][0]),cmul(S_down,temp[1][1]));
			uvert[count]=cadd(temp[1][3],uvert[count]);
			uvert[count]=rcmul((1.0/sfact),uvert[count]);
		}
		if(w < dw)w=0.0;	/* reset w */
		count++;
	}
	
		/* Urad(w) and Uvert(w) are now the response of the layered media due to */
		/* an SH impluse into the bottom of the stack */
		
		/* to calculate the ground displacement */
		/* inverse fourier transform the response */
		/* NOTE:  as shown in A&R equation 5.59 the */
		/* inverse Fourier transform to get ground */
		/* displacement has the opposite sign as the */
		/* convention used in A&R, therefore, we'll take the */
		/* complex conjugate before taking the inverse fft */
		
	for(i=0;i<nf;i++)
	{
		urad[i]=conjg(urad[i]);
		uvert[i]=conjg(uvert[i]);
	}
		
		/* to remove ringing in the time domain let's taper the */
		/* spectrum at high frequency.  Taper is a hardwired */
		/* cosine taper over the spectrum */
		
	if(taper == 'y')
	{
		for(i=0;i<=data_length/2;i++)
		{
			scale=(float)(i)/(float)(data_length/2);
			scale=cos( (scale*1.57079633) );
			urad[i]=rcmul(scale,urad[i]);
			uvert[i]=rcmul(scale,uvert[i]);
		}
	}
	cfftri((float *)urad,data_length);
	cfftri((float *)uvert,data_length);
	
	for(i=0;i<data_length/2;i++) /* get the correct amplitudes */
	{
		urad[i]=rcmul((2./(float)data_length),urad[i]);
		uvert[i]=rcmul((2./(float)data_length),uvert[i]);
	}
	return;
}

/*
	This routine calculates the ground displacement observed from an
	SH wave impinging from below a stack of layers

	Propagator matrices are used, Notation is as in Aki and Richards,
	Chapter 5.
	
	Need Subroutines;
	
		makeF_SH -> Calculates the Layer Matrix F for each layer at
					one frequency
		makeFinv_SH	-> Calculates the inverse of the Layer Matrix F for
						each layer at one frequency
		cmat_mult -> Multiplies two complex matrices (this calls routines
					to perform complex arithmetic
*/

void SHsurfacedisplace(float z[], float beta[], float rho[], float q[],
	int number_layers, float ray_p, complex u[], float nyquist,
	int data_length,char taper)
{
	complex mu[MAXLAYERS+1];
	complex eta[MAXLAYERS+1];
	
				/* beta, mu, eta, z define the properties of each layer */
				/* beta -> shear wave velocity */
				/*	mu -> shear modulus */
				/* z -> the depth to the bottom of each layer */
				/*		the depth to the last layer is not used */
				/*		last layer is assumed to be a half space */
				/* eta -> cosj/beta in each layer */
				/* NOTE all of these values start from the shallowest */
				/* layer to the deepest.  First layer is stored at */
				/* an index of 1.  The index zero is the free surface */
				
				/* u -> time domain displacment at the suface */
				/* ray_p -> ray parameter of incoming SH wave */
				/* nyquist -> nyquist frequency to be used */
				/* data_length -> length of the output time series */
				/*					must be a factor of 2 */

				
	complex F[2][2],Finv[2][2];	/* layer matrices See Aki and Richards */
	complex Sdown;	/* complex response of the reflected S */
	complex vel;	/* temporary array to hold complex velocity */
	complex temp[2][2],temp1[2][2],temp2[2][2],n2; /* work space */
	int i,j,count,nf,layer;
	float w,dw,freq_inc,zref,scale,sfact;

				/* Prototype all called functions */
	void makeF_SH(complex mu,complex eta,float zref,float z,float w,complex F[2][2]);
	void makeFinv_SH(complex mu,complex eta,float zref,float z,float w,complex Finv[2][2]);
	void cmat_mult(complex *F,int m, int n, complex *Finv,int p, complex *temp1);
	void cfftri(float *a,int dlength);	
	complex cdiv(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cmul(complex a,complex b);
	complex cadd(complex a,complex b);
	complex csqrt(complex a);
	complex conjg(complex a);
	float mscale(complex *f,int m, int n);	
	
		/* set reference depth to depth of bottom of the stack */
	
	zref=z[number_layers-1];
	z[0]=0.0;

		/* calculate frequency increment and the number of frequencies */
		
	nf=(int)(data_length/2.0+1.0);
	freq_inc=nyquist/(float)(nf-1.0);
	dw=freq_inc*2.0*PI;
	
			/* loop over each frequency and calculate ground response */
	count=0;
	for(w=0.0;w < nyquist*2.0*PI+dw;w+=dw)
	{
				/* reset at temporary storage array */
				
		temp[0][0].r=temp[1][1].r=1.0;
		temp[0][0].i=temp[1][1].i=temp[0][1].r=0.0;
		temp[0][1].i=temp[1][0].r=temp[1][0].i=0.0;	/* identity matrix */
		sfact=1.0;
		
		if(w == 0.0)w+=(dw/100.0);	/* set w to something small */
			
			/* calculate mu and eta for each layer */
		
		for(i=1;i<=number_layers;i++)
		{
			vel.r=beta[i]*(1.0+log(w/6.28318531)*1.0/(3.1415926*q[i]));
			vel.i= -1.0*beta[i]/(2.0*q[i]);
			mu[i]=cmul(vel,vel);
			mu[i]=rcmul((rho[i]*CONVRHO),mu[i]);
			
			n2=cmul(vel,vel);
			n2=cdiv(comp(1.0,0.0),n2);
			n2=cadd(comp((-1.0*ray_p*ray_p),0.0),n2);
			eta[i]=csqrt(n2);
		}
		
			/* construct propagator matrices for each layer */
						
		for(layer=0;layer < number_layers; layer++)
		{
				/* for every layer but the half space get F*Finv */
					
			if(layer != number_layers-1)
			{
				makeF_SH(mu[layer+1],eta[layer+1],zref,z[layer],w,F);
				makeFinv_SH(mu[layer+1],eta[layer+1],zref,z[layer+1],w,Finv);
				cmat_mult((complex *)F,2,2,(complex *)Finv,2,(complex *)temp1);
				/*sfact *= mscale((complex *)temp1,2,2);*/
					
						/* accumulate these over all layers */
							
				cmat_mult((complex *)temp,2,2,(complex *)temp1,2,(complex *)temp2);
				for(i=0;i<2;i++)
					for(j=0;j<2;j++)
						temp[i][j]=temp2[i][j];
			}
			else	/* in half space only calculate F */
			{
				makeF_SH(mu[layer+1],eta[layer+1],zref,z[layer],w,F);
				cmat_mult((complex *)temp,2,2,(complex *)F,2,(complex *)temp1);
				for(i=0;i<2;i++)
					for(j=0;j<2;j++)
						temp[i][j]=temp1[i][j];
			}
		}
				/* temp now contains the contribution of each layer */
				/* at frequency w */
				/* now solve the matrix equation */
				/*
					U		temp[0][0]	temp[0][1]		1
						=
					0		temp[1][0]	temp[1][1]		S-down
				*/
				/*	where U is the observed ground displacment */
				/* 	and S-down the the reflected S wave coming */
				/*	out of the bottom of the stack */
					
		Sdown=cdiv(temp[1][1],temp[1][0]);
		Sdown=rcmul((float)-1.0,Sdown);
		u[count]=cmul(temp[0][0],Sdown);
		u[count]=cadd(temp[0][1],u[count]);
		u[count]=rcmul((1.0/sfact),u[count]);
	
		if(w < dw)w=0.0;	/* reset w */
		count++;
	}
	
		/* U(w) is now the response of the layered media due to */
		/* an SH impluse into the bottom of the stack */
		
		/* to calculate the ground displacement */
		/* inverse fourier transform the response */
		/* NOTE:  as shown in A&R equation 5.59 the */
		/* inverse Fourier transform to get ground */
		/* displacement has the opposite sign as the */
		/* convention used in A&R, therefore, we'll take the */
		/* complex conjugate before taking the inverse fft */
		
	for(i=0;i<nf;i++)
		u[i]=conjg(u[i]);
		
		/* to remove ringing in the time domain let's taper the */
		/* spectrum at high frequency.  Taper is a hardwired */
		/* cosine taper over the spectrum */
		
	if(taper == 'y')
	{
		for(i=0;i<=data_length/2;i++)
		{
			scale=(float)(i)/(float)(data_length/2);
			scale=cos( (scale*1.57079633) );
			u[i].r*=scale;
			u[i].i*=scale;
		}
	}
	cfftri((float *)u,data_length);
	for(i=0;i<data_length/2;i++) /* get the correct amplitudes */
	{
		u[i].r*=(2./(float)data_length);
		u[i].i*=(2./(float)data_length);
	}
	return;
}


void makeF_PSV(float rho,float p,complex alpha,complex beta,complex xi,complex eta,float zref,float z,float w, 
	complex F[4][4])

/*
		This routine calculates the solution matrix for the PSV problem of
		a single layer at a single frequency. All notation is as in Aki
		and Richards, page 165.
*/

{
	complex E[4][4],GAMMA[4][4],exponent;
	
			/* Prototype all called functions */
	complex cmul(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cmul(complex a,complex b);
	complex cexp(complex a);
	complex comp(float r,float i);
	void cmat_mult(complex *a,int m,int n,complex *b,int p,complex *c);
	
		/* F is calculated via equation 5.63 and 5.65 */
		/* first set the submatrices E and GAMMA */
		
	E[0][0]=rcmul(p,alpha);
	E[0][1]=cmul(beta,eta);
	E[0][2]=E[0][0];
	E[0][3]=E[0][1];
	E[1][0]=cmul(alpha,xi);
	E[1][1]=rcmul((-1.0*p),beta);
	E[1][2]=rcmul(-1.0,E[1][0]);
	E[1][3]=rcmul(-1.0,E[1][1]);
	E[2][0]=cmul(beta,beta);
	E[2][0]=cmul(alpha,E[2][0]);
	E[2][0]=cmul(E[2][0],xi);
	E[2][0]=cmul(comp(0.0,(2.0*w*rho*p)),E[2][0]);
	E[2][1]=cmul(beta,beta);
	E[2][1]=rcmul((-2.0*p*p),E[2][1]);
	E[2][1]=cadd(comp(1.0,0.0),E[2][1]);
	E[2][1]=cmul(beta,E[2][1]);
	E[2][1]=cmul(comp(0.0,(w*rho)),E[2][1]);
	E[2][2]=rcmul(-1.0,E[2][0]);
	E[2][3]=rcmul(-1.0,E[2][1]);
	E[3][0]=cmul(beta,beta);
	E[3][0]=rcmul((-1.0*2.0*p*p),E[3][0]);
	E[3][0]=cadd(comp(1.0,0.0),E[3][0]);
	E[3][0]=cmul(alpha,E[3][0]);
	E[3][0]=cmul(comp(0.0,w*rho),E[3][0]);
	E[3][1]=cmul(beta,beta);
	E[3][1]=cmul(beta,E[3][1]);
	E[3][1]=cmul(eta,E[3][1]);
	E[3][1]=cmul(comp(0.0,(-2.0*w*rho*p)),E[3][1]);
	E[3][2]=E[3][0];
	E[3][3]=E[3][1];

	GAMMA[0][1]=GAMMA[0][2]=GAMMA[0][3]=GAMMA[1][0]=GAMMA[1][2]=comp(0.0,0.0);
	GAMMA[1][3]=GAMMA[2][0]=GAMMA[2][1]=GAMMA[2][3]=GAMMA[3][0]=comp(0.0,0.0);
	GAMMA[3][1]=GAMMA[3][2]=comp(0.0,0.0);
	exponent=cmul(comp(0.0,(w*(z-zref))),xi);
	GAMMA[0][0]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMA[2][2]=cexp(exponent);
	exponent=cmul(comp(0.0,(w*(z-zref))),eta);
	GAMMA[1][1]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMA[3][3]=cexp(exponent);
	
		/* F=E*GAMMA */
		
	cmat_mult((complex *)E,4,4,(complex *)GAMMA,4,(complex *)F);
}


void makeFinv_PSV(float rho,float p,complex alpha,complex beta,complex xi,complex eta,float zref,float z,float w, 
	complex Finv[4][4])

/*
		This routine calculates the inverse of the solution matrix for 
		the SH problem of a single layer at a single frequency. 
		All notation is as in Aki and Richards, page 166.
*/

{
	complex Einv[4][4],GAMMAinv[4][4],exponent,denominator;
	
		/* Prototype all called functions */
	complex comp(float r,float i);
	complex cmul(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cexp(complex a);
	complex cdiv(complex a,complex b);
	void cmat_mult(complex *a,int m,int n,complex *b,int p,complex *c);
	
		/* Finv is calculated via equation 5.64, 5.65, and 5.67 */
		/* first set the submatrices Einv and GAMMAinv */
		
	Einv[0][0]=cmul(beta,beta);
	Einv[0][0]=cdiv(Einv[0][0],alpha);
	Einv[0][0]=rcmul(p,Einv[0][0]);
	
	denominator=cmul(alpha,xi);
	denominator=rcmul(2.0,denominator);
	Einv[0][1]=cmul(beta,beta);
	Einv[0][1]=rcmul((-2.0*p*p),Einv[0][1]);
	Einv[0][1]=cadd(comp(1.0,0.0),Einv[0][1]);
	Einv[0][1]=cdiv(Einv[0][1],denominator);
	
	denominator=cmul(alpha,xi);
	denominator=rcmul((2.0*w*rho),denominator);
	Einv[0][2]=cdiv(comp(0.0,(-1.0*p)),denominator);
	
	denominator=rcmul((2.0*w*rho),alpha);
	Einv[0][3]=cdiv(comp(0.0,-1.0),denominator);
	
	denominator=cmul(beta,eta);
	denominator=rcmul(2.0,denominator);
	Einv[1][0]=cmul(beta,beta);
	Einv[1][0]=rcmul((-2.0*p*p),Einv[1][0]);
	Einv[1][0]=cadd(comp(1.0,0.0),Einv[1][0]);
	Einv[1][0]=cdiv(Einv[1][0],denominator);
	
	Einv[1][1]=rcmul((-1.0*p),beta);

	denominator=rcmul((2.0*w*rho),beta);
	Einv[1][2]=cdiv(comp(0.0,-1.0),denominator);
	
	denominator=cmul(beta,eta);
	denominator=rcmul((2.0*w*rho),denominator);
	Einv[1][3]=cdiv(comp(0.0,p),denominator);

	Einv[2][0]=Einv[0][0];
	Einv[2][1]=rcmul(-1.0,Einv[0][1]);
	Einv[2][2]=rcmul(-1.0,Einv[0][2]);
	Einv[2][3]=Einv[0][3];
	
	Einv[3][0]=Einv[1][0];
	Einv[3][1]=rcmul(-1.0,Einv[1][1]);
	Einv[3][2]=rcmul(-1.0,Einv[1][2]);
	Einv[3][3]=Einv[1][3];


	GAMMAinv[0][1]=GAMMAinv[0][2]=GAMMAinv[0][3]=GAMMAinv[1][0]=GAMMAinv[1][2]=comp(0.0,0.0);
	GAMMAinv[1][3]=GAMMAinv[2][0]=GAMMAinv[2][1]=GAMMAinv[2][3]=GAMMAinv[3][0]=comp(0.0,0.0);
	GAMMAinv[3][1]=GAMMAinv[3][2]=comp(0.0,0.0);
	exponent=cmul(comp(0.0,(-1.0*w*(z-zref))),xi);
	GAMMAinv[0][0]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMAinv[2][2]=cexp(exponent);
	exponent=cmul(comp(0.0,(-1.0*w*(z-zref))),eta);
	GAMMAinv[1][1]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMAinv[3][3]=cexp(exponent);
		
		/* F=E*GAMMA */
		
	cmat_mult((complex *)GAMMAinv,4,4,(complex *)Einv,4,(complex *)Finv);
}


void makeF_SH(complex mu,complex eta,float zref,float z,float w, complex F[2][2])

/*
		This routine calculates the solution matrix for the SH problem of
		a single layer at a single frequency. All notation is as in Aki
		and Richards, page 165.
*/

{
	complex E[2][2],GAMMA[2][2],exponent;
	
			/* Prototype all called functions */
	complex cmul(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cmul(complex a,complex b);
	complex cexp(complex a);
	complex comp(float r,float i);
	void cmat_mult(complex *a,int m,int n,complex *b,int p,complex *c);
	
		/* F is calculated via equation 5.63 */
		/* first set the submatrices E and GAMMA */
		
	E[0][0]=comp(1.0,0.0);
	E[0][1]=E[0][0];
	E[1][0]=cmul(mu,eta);
	E[1][0]=cmul(comp(0.0,1.0),E[1][0]);
	E[1][0]=rcmul(w,E[1][0]);
	E[1][1]=rcmul(-1.0,E[1][0]);
	
	GAMMA[0][1]=comp(0.0,0.0);
	GAMMA[1][0]=GAMMA[0][1];
	exponent=cmul(comp(0.0,w*(z-zref)), eta);
	GAMMA[0][0]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMA[1][1]=cexp(exponent);
	
		/* F=E*GAMMA */
		
	cmat_mult((complex *)E,2,2,(complex *)GAMMA,2,(complex *)F);
}


void makeFinv_SH(complex mu,complex eta,float zref,float z,float w, complex Finv[2][2])

/*
		This routine calculates the inverse of the solution matrix for 
		the SH problem of a single layer at a single frequency. 
		All notation is as in Aki and Richards, page 166.
*/

{
	complex Einv[2][2],GAMMAinv[2][2],exponent,denominator;
	
		/* Prototype all called functions */
	complex comp(float r,float i);
	complex cmul(complex a,complex b);
	complex rcmul(float x,complex a);
	complex cexp(complex a);
	complex cdiv(complex a,complex b);
	void cmat_mult(complex *a,int m,int n,complex *b,int p,complex *c);
	
		/* Finv is calculated via equation 5.64 */
		/* first set the submatrices Einv and GAMMAinv */
		
	Einv[0][0]=comp(0.5,0.0);
	Einv[1][0]=Einv[0][0];
	denominator=cmul(mu,eta);
	denominator=rcmul((2.0*w),denominator);
	Einv[1][1]=cdiv(comp(0.0,1.0), denominator);
	Einv[0][1]=rcmul(-1.0,Einv[1][1]);

	GAMMAinv[0][1]=comp(0.0,0.0);
	GAMMAinv[1][0]=GAMMAinv[0][1];
	exponent=cmul(comp(0.0,w*(z-zref)), eta);
	GAMMAinv[1][1]=cexp(exponent);
	exponent=rcmul(-1.0,exponent);
	GAMMAinv[0][0]=cexp(exponent);
		
		/* F=E*GAMMA */
		
	cmat_mult((complex *)GAMMAinv,2,2,(complex *)Einv,2,(complex *)Finv);
}


void cfftri( x, n )
int n;
float x[];
{
	int nn, is, nm, j, i, k1j, k1i, k2j, k2i;
	float s, fn, ex, wr, wi, wwr, wwi, wrr, a1, a2, b1, b2;

	void cfour(float x[], int nn, int is);
	nn = n/2;
	s = x[0];
	x[0] = 0.5 * ( x[0] + x[n] );
	x[1] = 0.5 * ( s - x[n] );
	x[nn+1] = (-x[nn+1]);
	is = -1;
	nm = nn/2;
	fn = (float) n ;
	ex = 6.2831853 / fn ;
	j = nn;
	wr = 1.0;
	wi = 0.0;
	wwr = (float) cos ( (double) ex );
	wwi = (float) ( - sin( (double) ex ) );
	for ( i=2; i<=nm; i++ ) {
		wrr = wr*wwr-wi*wwi;
		wi = wr*wwi+wi*wwr;
		wr = wrr;
		k1j = 2*j-1;
		k1i = 2*i-1;
		k2j = 2*j;
		k2i = 2*i;
		a1 = 0.5 * ( x[k1i-1] + x[k1j-1] );
		a2 = 0.5 * ( x[k2i-1] - x[k2j-1] );
		b1 = 0.5 * (-x[k1i-1] + x[k1j-1] );
		b2 = 0.5 * (-x[k2i-1] - x[k2j-1] );
		s = b1;
		b1 = b1*wr+b2*wi;
		b2 = b2*wr-s*wi;
		x[k1i-1] = (a1 - b2);
		x[k2i-1] = (-a2-b1);
		x[k1j-1] = (a1+b2);
		x[k2j-1] = (a2-b1);
		j -= 1;
		}
	cfour(x, nn, is);
	return;
	}
	
void cfour( data, n, isign )
int n, isign;
float data[];
{
	int ip0, ip1, ip2, ip3, i3rev;
	int i1, i2a, i2b, i3;
	float sinth, wstpr, wstpi, wr, wi, tempr, tempi, theta;
	double sin();
	ip0=2;
	ip3=ip0*n;
	i3rev=1;
	for( i3=1; i3<=ip3; i3+=ip0 ) {
		if( i3 < i3rev ) {
			tempr = data[i3-1];
			tempi = data[i3];
			data[i3-1] = data[i3rev-1];
			data[i3] = data[i3rev];
			data[i3rev-1] = tempr;
			data[i3rev] = tempi;
			}
		ip1 = ip3 / 2;
		do {
			if( i3rev <= ip1 )
				break;
			i3rev -= ip1;
			ip1 /= 2;
			}
			while ( ip1 >= ip0 );
		i3rev += ip1;
		}
		ip1 = ip0;
		while ( ip1 < ip3 ) {
			ip2 = ip1 * 2;
			theta = 6.283185/( (float) (isign*ip2/ip0) );
			sinth = (float) sin( (double) (theta/2.) );
			wstpr = -2.*sinth*sinth;
			wstpi = (float) sin( (double) theta );
			wr = 1.;
			wi = 0.;
			for ( i1=1; i1<=ip1; i1+=ip0 ) {
			for ( i3=i1; i3<ip3; i3+=ip2 ) {
				i2a=i3;
				i2b=i2a+ip1;
				tempr = wr*data[i2b-1] - wi*data[i2b];
				tempi = wr*data[i2b] + wi*data[i2b-1];
				data[i2b-1] = data[i2a-1] - tempr;
				data[i2b] = data[i2a] - tempi;
				data[i2a-1] += tempr;
				data[i2a] += tempi;
				}
				tempr = wr;
				wr = wr*wstpr - wi*wstpi + wr;
				wi = wi*wstpr + tempr*wstpi + wi;
				}
			ip1=ip2;
			}
		return;
		}

/* convolves two timeseries in the time domain
 *
 * Inputs:
 *	a,b=two timeseries to be convolved
 *	ndata_a,ndata_b=lengths of the timeseries
 *
 * Outputs:
 *	routine returns a pointer to an array ndata_a+ndata_b-1 points long
 *	containing the convolved trace.  
 *
 */

void t_fold(float *a,int ndata_a,float *b,int ndata_b,float *result)

{
	int i,j,k;

	for(i=0;i<ndata_a;++i)
	{
		for(j=0;j<ndata_b;++j)
		{
			k=i+j;
			result[k]=a[i]*b[j]+result[k];
		}
	}

	return;
}
